"""Aggregation of the precise parser and the byline parser,
leading to high-level and complete parsing.

"""

from clyngor import asp_parsing
from clyngor.asp_parsing import precise_parser
from clyngor.asp_parsing import byline_parser







from timeit import timeit
from pprint import pprint
from functools import partial

from clyngor import ASP
from clyngor import asp_parsing
from clyngor.asp_parsing import alt_parse


ASP_CODE = """
rel(a,c). rel(a,d). rel(b,d). rel(b,e).
obj(X):- rel(X,_) ; rel(X,Y): att(Y).
att(Y):- rel(_,Y) ; rel(X,Y): obj(X).
%* w.
- coucou
*%
:- not obj(X):obj(X).
:- not att(Y):att(Y).
"""

def time_efficiency(parser:callable, number:int=1000) -> float:
    def run():
        return asp_parsing.program_to_dependancy_graph(parser(ASP_CODE))
    return 'Perform 1000 graphs in {} seconds.'.format(timeit(run, number=number))

def time_efficiency_alt():
    program = (ASP_CODE)
    func = partial(asp_parsing.program_to_dependancy_graph, program, have_comments=False)
    return 'Perform 1000 graphs in ' + str(timeit(func, number=1000)) + ' seconds.'
    # last time: 0.01s


if __name__ == '__main__':
    answers = ASP(ASP_CODE)
    for answer in answers.by_predicate.first_arg_only:
        print('{' + ','.join(answer['obj']) + '} × {' + ','.join(answer['att']) + '}')
    print()

    print('Dependancy graph:')
    pprint(asp_parsing.program_to_dependancy_graph(ASP_CODE))
    print()

    print('Benchmark:')
    for parser in (asp_parsing.precise_parser.parse_asp_program_by_arpeggio,  # last time: 4.0s, 5.8s
                   asp_parsing.precise_parser.parse_asp_program_by_pypeg):  # last time: 8.5s, 9.5s
        print(parser.__name__.ljust(17) + ':', time_efficiency(parser))
    print()


    print('ALTERNATIVE:')
    program = alt_parse.parse_asp(ASP_CODE)
    print(asp_parsing.program_to_dependancy_graph(program))


    print('PRETTY PRINTING:')
    from pprint import pprint
    program = asp_parsing.parse_asp_program(ASP_CODE)
    pprint(program)
    print('\n'.join(asp_parsing.pprint.to_asp_source_code(program)))



"""Low level interface to clingo.

"""


import re
import tempfile
import subprocess
import clyngor
from clyngor.answers import Answers
from clyngor.utils import cleaned_path


def solve(files:iter=(), options:iter=[], inline:str=None,
          subproc_shell:bool=False, print_command:bool=False,
          nb_model:int=0, time_limit:int=0, constants:dict={},
          clean_path:bool=True) -> iter:
    """Run the solver on given files, with given options, and return
    an Answers instance yielding answer sets.

    files -- iterable of files feeding the solver
    options -- string or iterable of options for clingo
    inline -- ASP source code to feed the solver with
    subproc_shell -- use shell=True in subprocess call (NB: you should not)
    print_command -- print full command to stdout before running it
    clean_path -- clean the path of given files before using them

    Shortcut to clingo's options:
    nb_model -- number of model to output (0 for all (default))
    time_limit -- zero or number of seconds to wait before interrupting solving
    constants -- mapping name -> value of constants for the grounding

    """
    files = tuple(map(cleaned_path, files) if clean_path else files)
    run_command = command(files, options, inline, nb_model, time_limit, constants)
    if print_command:
        print(run_command)

    clingo = subprocess.Popen(
        run_command,
        stderr = subprocess.PIPE,
        stdout = subprocess.PIPE,
        shell=bool(subproc_shell),
    )

    def gen_answers():
        stdout = iter(clingo.stdout)
        while True:
            cur_line = next(stdout).decode()
            print('LINE?:', cur_line)
            if cur_line.startswith('Answer: '):
                model = next(stdout).decode()
                print('MODEL?:', model)
                yield model

    return Answers(gen_answers(), command=' '.join(run_command))


def command(files:iter=(), options:iter=[], inline:str=None,
            nb_model:int=0, time_limit:int=0, constants:dict={}) -> iter:
    """Return the shell command running the solver on given files,
    with given options.

    files -- iterable of files feeding the solver
    options -- string or iterable of options for clingo
    inline -- ASP source code to feed the solver with

    Shortcut to clingo's options:
    nb_model -- number of model to output (0 for all (default))
    time_limit -- zero or number of seconds to wait before interrupting solving
    constants -- mapping name -> value of constants for the grounding

    """
    files = [files] if isinstance(files, str) else list(files)
    options = [options] if isinstance(options, str) else list(options)
    if inline:
        with tempfile.NamedTemporaryFile(mode='w', delete=False) as fd:
            fd.write(inline)
            files = tuple(files) + (fd.name,)
    if constants:
        for name, value in constants.items():
            options.append('-c {}={}'.format(name,value))
    if time_limit:
        try:
            time_limit = int(time_limit)
        except ValueError:
            raise ValueError("Time limit should be integer, not " + type(time_limit).__name__)
        options.append('--time-limit=' + str(time_limit))
    if nb_model:
        try:
            nb_model = int(nb_model)
        except ValueError:
            raise ValueError("Number of model must be int, not " + type(nb_model).__name__)
        if nb_model < 0:
            raise ValueError("Number of model must be >= 0.")
    else:
        nb_model = 0

    return [clyngor.CLINGO_BIN_PATH, *options, *files, '-n ' + str(nb_model)]


def clingo_version() -> dict:
    """Return clingo's version information in a dict"""
    clingo = subprocess.Popen(
        [clyngor.CLINGO_BIN_PATH, '--version', '--outf=2'],
        stderr = subprocess.PIPE,
        stdout = subprocess.PIPE,
    )
    fields = {
        'address model': re.compile(r'Address model: ([3264]{2})-bit'),
        'clingo version': re.compile(r'clingo version ([0-9\.]+)'),
        'libgringo': re.compile(r'libgringo version ([0-9\.]+)'),
        'libclasp': re.compile(r'libclasp version ([0-9\.]+)'),
        'libpotassco': re.compile(r'libpotassco version ([0-9\.]+)'),
        'python': re.compile(r'with[out]{0,3}\sPython\s?([0-9\.]+)?'),
        'lua': re.compile(r'with[out]{0,3}\sLua\s?([0-9\.]+)?'),
    }
    stdout = clingo.communicate()[0].decode()

    return {
        field: reg.search(stdout).groups()[0]
        for field, reg in fields.items()
    }



"""Routines working on ASP source in order to parse them, line by line instead
of a whole.

"""

import math
import itertools
from bisect import bisect_left
from collections import defaultdict, namedtuple
import arpeggio as ap


SourceBlock = namedtuple('SourceBlock', 'have_code, lines, start, end')
def SourceBlock__from_raw(have_code, lines, start, end):
    return SourceBlock(bool(have_code), tuple(lines), int(start), int(end))
SourceBlock.from_raw = SourceBlock__from_raw


def parse(asp_source_code:str) -> iter:
    """Parse the source code into hierarchy of lines"""
    parser = ap.ParserPython(line_grammar())
    parse_tree = parser.parse(asp_source_code)
    parsed_lines = ap.visit_parse_tree(parse_tree, LinesExtractor())
    return with_line_number(parsed_lines, asp_source_code)


def clusterize_from_source(source_code:str) -> tuple:
    """Return the clusterized source code, parsed into blocks"""
    return tuple(clusterize(parse(source_code)))


def rebuild_clusters(clusters:str, sep_cluster:str='\n\n', sep_comment:str='  ',
                     rule_indent:str='    ') -> str:
    """Return the source code inferred from given clusters.

    """
    return sep_cluster.join(cluster_as_string(cluster, sep_comment, rule_indent)
                            for cluster in clusters)


def cluster_as_string(cluster:SourceBlock, sep_comment:str, rule_indent:str) -> str:
    """Return a string representation of given cluster"""
    COMMENTS = {'comment', 'multicomment'}
    CODE = {'code', 'end'}
    pretty_doc = lambda cluster: '\n'.join(line[4] for line in cluster.lines)
    previous_line_end = None
    previous_type = None
    out = ''
    in_code = False

    def spacing() -> str:
        if not new_line:
            if previous_type in COMMENTS:
                return sep_comment
            elif previous_type in CODE and type in COMMENTS:
                return sep_comment
        elif new_line and not first_line:
            if in_code:  # put indentation
                return '\n' + rule_indent
            else:
                return '\n'
        return ''

    for line_start, line_end, type, _, data in cluster.lines:
        assert line_end is not None
        assert line_start is not None
        first_line = previous_line_end is None
        new_line = first_line or line_start != previous_line_end
        out += spacing()
        if type == 'comment':
            out += '%' + data
        if type == 'multicomment':
            out += '%*' + data + '*%'
        if type == 'code':
            in_code = True
            out += data
        if type == 'end':
            if not in_code:  # if in_code is True, there is a problem
                print("MISUTD: unexpected '.' during parsing of non-code segment")
            in_code = False
            out += data
        if type == 'text':
            out += '"' + data + '"'
        previous_line_end = line_end
        previous_type = type
    return out


def clusterize(parsed_lines:iter) -> [SourceBlock]:
    """Yield SourceBlock instances.

    """
    line_type = lambda line: line[2]
    have_code = lambda line: 'code' == line_type(line)
    have_end = lambda line: 'end' == line_type(line)
    empty = lambda line: 'empty' == line_type(line)

    state_code = False
    code_in_cluster, current_cluster, start, end = False, [], math.inf, 0

    for line in parsed_lines:
        if empty(line):
            if not state_code and current_cluster:
                yield SourceBlock.from_raw(code_in_cluster, current_cluster, start, end)
                code_in_cluster, current_cluster, start, end = False, [], math.inf, 0
            continue
        start = min(start, line[0])
        end = max(end, line[1])
        if state_code and not have_end(line):
            current_cluster.append(line)
        elif have_code(line):
            assert not have_end(line)
            state_code = True
            code_in_cluster = True
            current_cluster.append(line)
        elif have_end(line):
            assert not have_code(line)
            assert state_code, "input source is not valid ; dot arrives before rules"
            state_code = False
            current_cluster.append(line)
        else:  # line is just free doc
            current_cluster.append(line)
    if current_cluster:
        yield SourceBlock.from_raw(code_in_cluster, current_cluster, start, end)


def build_line_index(parsed_lines:iter) -> dict:
    """Maps each line number with its content"""
    index = defaultdict(list)
    for full_line in parsed_lines:
        line_start, line_end, *data = full_line
        for line in range(line_start, line_end + 1):
            index[line].append(full_line)
    for line in range(max(index.keys())):  # create empty lines
        index[line]
    return {line: tuple(data) for line, data in index.items()}


def with_line_number(parsed_lines:iter, source:str) -> iter:
    lines = tuple(lines_end(source))
    line_of = lambda position: bisect_left(lines, position)
    previous_line_end = 0
    for line in parsed_lines:
        start_pos, end_pos = map(line_of, line[1])
        if start_pos > previous_line_end:
            yield previous_line_end, start_pos-1, 'empty'
        yield (start_pos, end_pos, *line)
        previous_line_end = end_pos+1


def lines_end(source:str) -> list:
    """Yield the position of each line end in source

    >>> tuple(lines_end('\\n'))
    (0,)
    >>> tuple(lines_end('\\na\\n'))
    (0, 2)
    >>> tuple(lines_end('\\na\\nb\\ncde\\n\\n'))
    (0, 2, 4, 8, 9)

    """
    size = 0
    for line in source.splitlines(True):
        yield len(line) + size - 1
        size += len(line)


def line_grammar():
    """Implement the Answer Set Programming grammar.

    """
    def text():         return ap.Sequence('"', ap.RegExMatch(r'((\\")|([^"]))*'), '"', skipws=False)
    def comment():      return ap.RegExMatch(r'%.*$'),
    def multiline_comment(): return ap.RegExMatch(r'%\*.*?\*%', multiline=True),
    def asp_code():     return ap.RegExMatch(r'[^%"\.]*[^%"\.\s]', multiline=True),
    def rule_end():     return '.',
    def program():      return ap.OneOrMore([text, multiline_comment, comment, asp_code, rule_end])

    return program


class LinesExtractor(ap.PTNodeVisitor):
    def __init__(self):
        super().__init__()

    def visit_text(self, node, children):
        # print(node.name, dir(node))
        # print(node.position, node.position_end, node.comments)
        # print(children)
        return 'text', (node.position, node.position_end), children[0]

    def visit_comment(self, node, children):
        assert children[0].startswith('%')
        return 'comment', (node.position, node.position_end), children[0][1:]

    def visit_multiline_comment(self, node, children):
        assert len(children) == 1
        assert children[0].endswith('*%')
        assert children[0].startswith('%*')
        assert len(children[0]) >= 4
        data = children[0][2:-2]
        return 'multicomment', (node.position, node.position_end), data

    def visit_asp_code(self, node, children):
        # data = children[0].strip('\n')
        # return 'code', (node.position, node.position + len(data) - 1), data
        return 'code', (node.position, node.position_end), children[0]

    def visit_rule_end(self, node, children):
        return 'end', (node.position, node.position_end), '.'

    def visit_program(self, node, children):
        return tuple(children)



"""Routines for answer set parsing.

"""
import re


import arpeggio as ap


class CollapsableAtomVisitor(ap.PTNodeVisitor):
    """Implement both the grammar and the way to handle it, dedicated to the
    parsing of ASP like string to produce frozenset instances.

    This is a more complex version than AtomVisitor, because implementing
    special use cases, notably partial cut of parsed atoms.

    collapse_args: function terms in predicate arguments are collapsed into strings
    collapse_atoms: atoms (predicate plus terms) are collapsed into strings
                   requires that collapse_args is True

    See Parser class for usage examples.

    """
    def __init__(self, collapse_args:bool=True, collapse_atoms:bool=False,
                 parse_integer:bool=True):
        super().__init__()
        self.collapse_args = bool(collapse_args)
        self.collapse_atoms = bool(collapse_atoms)
        self._int_builder = int if parse_integer else str

    def visit_number(self, node, children):
        return self._int_builder(node.value)

    def visit_args(self, node, children):
        return children

    def visit_text(self, node, children):
        text = tuple(children)[0]
        return '"' + text + '"'

    def visit_subterm(self, node, children):
        predicate, *args = children
        if self.collapse_args:
            return (predicate + '(' + ','.join(map(str, *args)) + ')') if args else predicate
        else:
            return (predicate, tuple(args[0])) if args else predicate

    def visit_term(self, node, children):
        predicate, *args = children
        if self.collapse_atoms:
            return (predicate + '(' + ','.join(map(str, *args)) + ')') if args else predicate
        else:
            return (predicate, (tuple(args[0]) if args else ()))

    def visit_terms(self, node, children):
        return frozenset(children)

    @staticmethod
    def grammar():
        def ident():      return ap.RegExMatch(r'[a-z][a-zA-Z0-9_]*')
        def number():     return ap.RegExMatch(r'-?[0-9]+')
        def text():       return '"', ap.RegExMatch(r'((\\")|([^"]))*'), '"'
        def litteral():   return [text, number]
        def subterm():    return [(ident, ap.Optional("(", args, ")")), litteral]
        def args():       return subterm, ap.ZeroOrMore(',', subterm)
        # NB: litteral outputed by #show are not handled.
        def term():       return ident, ap.Optional("(", args, ")")
        def terms():      return ap.ZeroOrMore(term)
        return terms


class AtomVisitor(ap.PTNodeVisitor):
    """Implement both the grammar and the way to handle it, dedicated to the
    parsing of ASP like string to produce frozenset instances.

    This is a simpler version than CollapsableAtomVisitor, which implement
    special use cases, notably partial cut of parsed atoms.

    """
    def __init__(self, parse_integer:bool=True):
        super().__init__()
        self._int_builder = int if parse_integer else str

    def visit_number(self, node, children):
        return self._int_builder(node.value)

    def visit_args(self, node, children):
        return children

    def visit_text(self, node, children):
        text = tuple(children)[0]
        return '"' + text + '"'

    def visit_term(self, node, children):
        predicate, *args = children
        return (predicate, tuple(args[0])) if args else predicate

    def visit_terms(self, node, children):
        return frozenset(children)

    @staticmethod
    def grammar():
        def ident():      return ap.RegExMatch(r'[a-z][a-zA-Z0-9_]*')
        def number():     return ap.RegExMatch(r'-?[0-9]+')
        def text():       return ap.ZeroOrMore([r'\"', ap.RegExMatch(r'[^"]*')])
        def litteral():   return [('"', text, '"'), number]
        def args():       return term, ap.ZeroOrMore(',', term)
        def term():       return [(ident, ap.Optional("(", args, ")")), litteral]
        def terms():      return ap.ZeroOrMore(term)
        return terms


class Parser:
    def __init__(self, collapse_atoms=False, collapse_args=True, callback=None,
                 parse_integer:bool=True):
        """
        collapse_args -- function terms in predicate arguments are collapsed into strings
        collapse_atoms -- atoms (predicate plus terms) are collapsed into strings
                          requires that collapse_args is True
        parse_integer -- return integers as int instead of string

        examples:

            >>> Parser(False, True).parse_terms('a(b,c(d))')
            frozenset({('a', ('b', 'c(d)'))})

            >>> Parser(True, True).parse_terms('a(b,c(d))')
            frozenset({'a(b,c(d))'})

            >>> Parser(False, False).parse_terms('a(b,c(d))')
            frozenset({('a', ('b', ('c', ('d',))))})

            >>> Parser(True, False).parse_terms('a(b,c(d))')  # doctest: +IGNORE_EXCEPTION_DETAIL
            Traceback (most recent call last):
            ...
            ValueError

            >>> Parser(parse_integer=False).parse_terms('a(3)')
            frozenset({('a', ('3',))})

            >>> Parser().parse_terms('a(3)')
            frozenset({('a', (3,))})

        """
        self.collapse_args = bool(collapse_args)
        self.collapse_atoms = bool(collapse_atoms)
        if not self.collapse_args and not self.collapse_atoms:  # optimized case
            self.atom_visitor = AtomVisitor(parse_integer)
        else:
            self.atom_visitor = CollapsableAtomVisitor(
                bool(collapse_args),
                bool(collapse_atoms),
                parse_integer
            )
        self.grammar = self.atom_visitor.grammar()
        self.callback = callback
        if self.collapse_atoms and not self.collapse_args:
            raise ValueError("if atoms are collapsed, terms must"
                             " also be collapsed!")


    def parse_terms(self, string:str) -> frozenset:
        """Return the frozenset computed from given valid ASP-compliant string"""
        parse_tree = ap.ParserPython(self.grammar).parse(string)
        if parse_tree:
            return ap.visit_parse_tree(parse_tree, self.atom_visitor)
        else:
            return frozenset()

    # alias
    parse = parse_terms


    def parse_clasp_output(self, output:iter or str, *, yield_stats:bool=False,
                           yield_info:bool=False):
        """Yield pairs (payload type, payload) where type is 'info', 'statistics'
        or 'answer' and payload the raw information.

        output -- iterable of lines or full clasp output to parse
        yield_stats -- yields final statistics as a mapping {field: value}
                       under type 'statistics'
        yield_info  -- yields first lines not related to first answer
                       under type 'info' as a tuple of lines

        In any case, tuple ('answer', termset) will be returned
        with termset a frozenset instance containing all atoms in the found model.

        See test/test_parsing.py for examples.

        """
        REGEX_ANSWER_HEADER = re.compile(r"Answer: [0-9]+")
        output = iter(output.splitlines() if isinstance(output, str) else output)
        modes = iter(('info', 'answers', 'statistics'))
        mode = next(modes)

        # all lines until meeting "Answer: 1" belongs to the info lines
        info = []
        line = ''
        line = next(output)
        while not REGEX_ANSWER_HEADER.fullmatch(line):
            info.append(line)
            line = next(output)
        if yield_info:
            yield 'info', tuple(info)

        # first answer begins
        while True:
            if REGEX_ANSWER_HEADER.fullmatch(line):
                next_line = next(output)
                answer = self.parse_terms(next_line)
                yield 'answer', answer
            if not line.strip():  # empty line: statistics are beginning
                if not yield_stats: break  # stats are the last part of the output
                stats = {}
                for line in output:
                    sep = line.find(':')
                    key, value = line[:sep], line[sep+1:]
                    stats[key.strip()] = value.strip()
                yield 'statistics', stats
            line = next(output)




from clyngor import ASP
from clyngor.asp_parsing import byline_parser
from clyngor.asp_parsing.byline_parser import SourceBlock


AWFUL_SOURCE = """
% this is a header
%*
a multiline one
%%*
*% % should be saved

%* this is a comment, inlined with a rule *% rel(a,(b;c;d;e)). %* *% rel(b,c) %

    .


% a comment before a rule
p(X):- q(X) ; not rel(a,X).

%* a multiline comment
before a rule
*%
p(X):- q(X,"%* text *% text %*") ; not rel(a,X).


p(X):- q(X) % a comment inside a rule
    ; not %* *% rel(a %*,Y).*% , X)
    .

% metaprogramming
#show. % do not show anything
% except…
#show p/1.
#show rel/2.
"""


def test_solving():
    """Prove that the program is valid"""
    answers = tuple(ASP(AWFUL_SOURCE).by_predicate)
    assert len(answers) == 1
    for idx, answer in enumerate(answers):
        assert 'p' not in answer
        assert len(answer['rel']) == 5
        assert answer['rel'] == {('a', 'b'), ('a', 'c'), ('a', 'd'), ('a', 'e'), ('b', 'c')}


def test_byline_rebuild():
    clusters = byline_parser.clusterize_from_source(AWFUL_SOURCE)
    rebuilt = byline_parser.rebuild_clusters(clusters, sep_comment=' ')
    lines = AWFUL_SOURCE.splitlines()
    # not all lines are conserved. Blank ones are discarded by rebuilding
    expected = lines[:8] + lines[9:10] + lines[11:19] + lines[20:]
    assert rebuilt == '\n'.join(expected).strip()


def test_byline_clusterize():
    clusters = tuple(byline_parser.clusterize(byline_parser.parse(AWFUL_SOURCE)))

    index, last_line_index = 1, None
    def entry(start, end, type, content, spacing:int=0) -> tuple:
        """Produce a line to put in a cluster. Automatically count the position"""
        nonlocal index, last_line_index
        if last_line_index is None:
            last_line_index = start
        types_size = {'comment': 1, 'multicomment': 4, 'code': 0, 'end': 0, 'text': 2}
        index += spacing + (start - last_line_index)  # take account of line break
        start_index = index
        index += len(content) + types_size[type]
        last_line_index = end
        return (start, end, type, (start_index, index), content)

    assert len(clusters) == 6
    assert clusters[0] == SourceBlock(
        False, (
            entry(1, 1, 'comment', ' this is a header'),
            entry(2, 5, 'multicomment', '\na multiline one\n%%*\n'),
            entry(5, 5, 'comment', ' should be saved', spacing=1)
        ), start=1, end=5
    )
    assert clusters[1] == SourceBlock(
        True, (
            entry(7, 7, 'multicomment', ' this is a comment, inlined with a rule '),
            entry(7, 7, 'code', 'rel(a,(b;c;d;e))', spacing=1),
            entry(7, 7, 'end', '.'),
            entry(7, 7, 'multicomment', ' ', spacing=1),
            entry(7, 7, 'code', 'rel(b,c)', spacing=1),
            entry(7, 7, 'comment', '', spacing=1),
            entry(9, 9, 'end', '.', spacing=4)
        ), start=7, end=9
    )
    assert clusters[2] == SourceBlock(
        True, (
            entry(12, 12, 'comment', ' a comment before a rule'),
            entry(13, 13, 'code', 'p(X):- q(X) ; not rel(a,X)'),
            entry(13, 13, 'end', '.')
        ), start=12, end=13
    )
    assert clusters[3] == SourceBlock(
        True, (
            entry(15, 17, 'multicomment', ' a multiline comment\nbefore a rule\n'),
            entry(18, 18, 'code', 'p(X):- q(X,'),
            entry(18, 18, 'text', '%* text *% text %*'),
            entry(18, 18, 'code', ') ; not rel(a,X)'),
            entry(18, 18, 'end', '.')
        ), start=15, end=18
    )
    assert clusters[4] == SourceBlock(
        True, (
            entry(21, 21, 'code', 'p(X):- q(X)'),
            entry(21, 21, 'comment', ' a comment inside a rule', spacing=1),
            entry(22, 22, 'code', '; not', spacing=4),
            entry(22, 22, 'multicomment', ' ', spacing=1),
            entry(22, 22, 'code', 'rel(a', spacing=1),
            entry(22, 22, 'multicomment', ',Y).', spacing=1),
            entry(22, 22, 'code', ', X)', spacing=1),
            entry(23, 23, 'end', '.', spacing=4)
        ), start=21, end=23
    )
    assert clusters[5] == SourceBlock(
        True, (
            entry(25, 25, 'comment', ' metaprogramming'),
            entry(26, 26, 'code', '#show'),
            entry(26, 26, 'end', '.'),
            entry(26, 26, 'comment', ' do not show anything', spacing=1),
            entry(27, 27, 'comment', ' except…'),
            entry(28, 28, 'code', '#show p/1'),
            entry(28, 28, 'end', '.'),
            entry(29, 29, 'code', '#show rel/2'),
            entry(29, 29, 'end', '.')
        ), start=25, end=29
    )




"""Definition of the ASP grammar"""


import arpeggio as ap


def asp_grammar():
    """Implement the Answer Set Programming grammar.

    """
    # litterals
    def ident():        return ap.RegExMatch(r'[a-z][a-zA-Z0-9_]*')
    def number():       return ap.RegExMatch(r'-?[0-9]+'),
    def text():         return '"', ap.RegExMatch(r'((\\")|([^"]))*'), '"'
    def variable():     return ap.RegExMatch(r'(([A-Z][a-zA-Z0-9_]*)|(_))')
    def litteral():     return [ident, variable, text, number]

    # maths
    def operator():     return list('+-/*\\') + ['**']
    def binop():        return [mathexp, ('(', mathexp, ')')], operator, [mathexp, ('(', mathexp, ')')]
    def mathexp():      return [number, binop]
    def assignment():   return variable, '=', mathexp

    # second level constructions
    def arg():          return [term, litteral, mathexp]
    def args():         return arg, ap.ZeroOrMore(',', arg)
    def multargs():     return args, ap.ZeroOrMore(';', args)

    # heads constructions
    def unnamedterm():  return '(', multargs, ')'
    def namedterm():    return ident, ap.Optional('(', multargs, ')')
    def term():         return [unnamedterm, namedterm]
    def selection():    return ap.Optional(number), '{', ap.OneOrMore(expression), '}', ap.Optional(number)

    # body constructions
    def not_term():     return 'not', namedterm
    def forall():       return namedterm, ':', term, ap.ZeroOrMore(',', term)
    def not_forall():   return 'not', namedterm, ':', term, ap.ZeroOrMore(',', term)
    def expression():   return [not_forall, forall, not_term, term, assignment]

    # metaprogramming
    def arityterm():    return ident, '/', ap.RegExMatch(r'[0-9]+')
    def meta_show():    return 'show', ap.Optional([forall, arityterm, term, litteral])
    def meta_const():   return 'const', ident, '=', litteral
    # def meta_max():   return 'const', ident, '=', litteral
    # def meta_min():   return 'const', ident, '=', litteral

    # program level constructions
    def body():         return expression, ap.ZeroOrMore(';', expression)
    def head():         return [selection, namedterm]
    def constraint():   return ':-', body
    def rule():         return head, ':-', body
    def multirule():    return (namedterm, ap.OneOrMore(';', namedterm)), ':-', body
    def meta():         return '#', [meta_show, meta_const]
    def program():      return ap.OneOrMore([meta, constraint, multirule, rule, head], '.')

    return program


def asp_grammar_comments():
    # return [ap.RegExMatch(r'%\*.*?\*%'), ap.RegExMatch(r'%.*$')]  # TODO: handle multiline comments
    # wait for Arpeggio's PR#38, which fix the problem
    # return [ap.RegExMatch(r'%\*.*?\*%', multiline=True), ap.RegExMatch(r'%.*$')]
    # Follows a non-efficient but working way to do multiline
    return [ap.RegExMatch(r'%\*((([^\*]?%)|([^%]?\*))|([^\*%]))*\*%'), ap.RegExMatch(r'%.*$')]





from clyngor.asp_parsing.precise_parser import parse_asp_program, CodeAsTuple


def test_const():
    ASP_CODE = r"""
    %follows a show
    #const a=2.
    #const ident="".
    #const ident="text$%\"".
    """
    rules = tuple(parse_asp_program(ASP_CODE, do=CodeAsTuple()))
    assert len(rules) == 3
    assert rules[0] == ('const', 'a', 2)
    assert rules[1] == ('const', 'ident', ('text', ''))
    assert rules[2] == ('const', 'ident', ('text', 'text$%\\"'))


def test_const_text_with_inlined_comment():
    ASP_CODE = r"""
    #const ident="%*text\" *%".
    """
    rules = tuple(parse_asp_program(ASP_CODE, do=CodeAsTuple()))
    assert len(rules) == 1
    assert rules[0] == ('const', 'ident', ('text', '')), (
        "if this fail, it's because comments are correctly handled.")


def test_show():
    ASP_CODE = r"""
    %follows a show
    #show a.
    #show a(X): b(X).
    #show a/1.
    """
    rules = tuple(parse_asp_program(ASP_CODE, do=CodeAsTuple()))
    assert len(rules) == 3
    assert rules[0] == ('show', 'a', ())
    assert rules[1] == ('show', 'a', ('X',), (('term', 'b', ('X',)),))
    assert rules[2] == ('show/n', 'a', 1)


def test_comments():
    ASP_CODE = r"""
    %a.
    %.
    %{}
    %a:- test
    %    *%
    % *
    ok.
    """
    rules = tuple(parse_asp_program(ASP_CODE, do=CodeAsTuple()))
    assert len(rules) == 1
    assert rules[0] == ('term', 'ok', ())


def test_non_working_multiline_comments():
    ASP_CODE = r"""
    %    *%
    %*
    ai.
    b: %bug
    :c- *%
    % *
    ok.
    """
    rules = tuple(parse_asp_program(ASP_CODE, do=CodeAsTuple()))
    assert len(rules) == 1
    assert rules[0] == ('term', 'ok', ()), "Multiline comments are not handled !"



def test_disjunction():
    ASP_CODE = r"""
    rel(a,(c;d)).
    rel(b,(d;e)).
    rel(c,(e(1),f;g,h(2))).
    """
    rules = tuple(parse_asp_program(ASP_CODE, do=CodeAsTuple()))
    assert len(rules) == 3
    first, second, third = rules

    print('FIRST:', first)
    expected = ('term', 'rel', (('term', 'a', ()),
                                ('disjunction',
                                 (('term', 'c', ()),),
                                 (('term', 'd', ()),)
                                )))
    print('EXPEC:', expected)
    assert first == expected

    print('SECOD:', second)
    expected = ('term', 'rel', (('term', 'b', ()),
                                ('disjunction',
                                 (('term', 'd', ()),),
                                 (('term', 'e', ()),)
                                )))
    print('EXPEC:', expected)
    assert second == expected
 
    print('THIRD:', third)
    expected = ('term', 'rel', (('term', 'c', ()),
                                ('disjunction',
                                 (('term', 'e', (1,)), ('term', 'f', ())),
                                 (('term', 'g', ()), ('term', 'h', (2,)))
                                )))
    print('EXPEC:', expected)
    assert third == expected


def test_constraint():
    ASP_CODE = r"""
    :- a.
    :- not obj(X):obj(X).
    """
    rules = tuple(parse_asp_program(ASP_CODE, do=CodeAsTuple()))
    assert len(rules) == 2
    first, second = rules

    print('FIRST:', first)
    expected = ('constraint', (('term', 'a', ()),) )
    print('EXPEC:', expected)
    assert first == expected

    print('SECOD:', second)
    expected = ('constraint', (('¬forall', 'obj', ('X',), (('term', 'obj', ('X',)),) ),) )
    print('EXPEC:', expected)
    assert second == expected


def test_forall():
    ASP_CODE = r"""
    a:- rel(X,Y): obj(X), att(Y).
    b:- not rel(X,Y): obj(X), att(Y).
    c:- not rel(X,Y): obj(X), att(Y) ; rel(c,_).
    """
    rules = tuple(parse_asp_program(ASP_CODE, do=CodeAsTuple()))
    assert len(rules) == 3
    first, second, third = rules

    print('FIRST:', first)
    expected = ('rule', ('term', 'a', ()), (('forall', 'rel', ('X', 'Y'), (('term', 'obj', ('X',)), ('term', 'att', ('Y',))) ),) )
    print('EXPEC:', expected)
    assert first == expected

    print('SECOD:', second)
    expected = ('rule', ('term', 'b', ()), (('¬forall', 'rel', ('X', 'Y'), (('term', 'obj', ('X',)), ('term', 'att', ('Y',))) ),) )
    print('EXPEC:', expected)
    assert second == expected

    print('THIRD:', third)
    expected = ('rule', ('term', 'c', ()), (('¬forall', 'rel', ('X', 'Y'), (('term', 'obj', ('X',)), ('term', 'att', ('Y',))) ), ('term', 'rel', (('term', 'c', ()), '_'))) )
    print('EXPEC:', expected)
    assert third == expected


def test_term():
    ASP_CODE = r"""
    a.
    b(1).
    c(2):- not b(2).
    """
    rules = tuple(parse_asp_program(ASP_CODE, do=CodeAsTuple()))
    assert len(rules) == 3
    first, second, third = rules

    print('FIRST:', first)
    expected = ('term', 'a', ())
    print('EXPEC:', expected)
    assert first == expected

    print('SECOD:', second)
    expected = ('term', 'b', (1,))
    print('EXPEC:', expected)
    assert second == expected

    print('THIRD:', third)
    expected = ('rule', ('term', 'c', (2,)), (('¬term', 'b', (2,)),) )
    print('EXPEC:', expected)
    assert third == expected


def test_selection():
    ASP_CODE = r"""
    1 { sel(X): obj(X) } 1.
    { con(X,Y): obj(X), att(Y, 2) }.
    2 { sel(X): obj(X) } 4:- anatom.
    """
    rules = tuple(parse_asp_program(ASP_CODE, do=CodeAsTuple()))
    assert len(rules) == 3

    first, second, third = rules

    print('FIRST:', first)
    expected = ('selection', 1, 1, (('forall', 'sel', ('X',), (('term', 'obj', ('X',)),)),),)
    print('EXPEC:', expected)
    assert first == expected

    print('SECOD:', second)
    expected = ('selection', 0, None, (('forall', 'con', ('X', 'Y'),
                                        (('term', 'obj', ('X',)),
                                         ('term', 'att', ('Y', 2)),),
                                        ),),)
    print('EXPEC:', expected)
    assert second == expected

    print('THIRD:', third)
    expected = ('rule',
                ('selection', 2, 4, (('forall', 'sel', ('X',),
                                     (('term', 'obj', ('X',)),),),)),
                (('term', 'anatom', ()),))
    print('EXPEC:', expected)
    assert third == expected


def test_multirules():
    ASP_CODE = r"""
    a;b:-c.
    b("string") ; a(X):- c(X).
    """

    rules = tuple(parse_asp_program(ASP_CODE))
    assert len(rules) == 2
    assert rules[0] == ('multirule', (
        ('term', 'a', ()),
        ('term', 'b', ()),
    ), (
        ('term', 'c', ()),
    ))
    assert rules[1] == ('multirule', (
        ('term', 'b', (('text', 'string'),)),
        ('term', 'a', ('X',)),
    ), (
        ('term', 'c', ('X',)),
    ))


def test_rules():
    ASP_CODE = r"""
    a.
    b("les amis, \"coucou\".").
    obj(X):- rel(X,_) ; rel(X,Y): att(Y).
    att(Y):- rel(_,Y) ; rel(X,Y): obj(X).
    """

    rules = tuple(parse_asp_program(ASP_CODE))
    assert len(rules) == 4

    first, second, third, fourth = rules
    assert first == ('term', 'a', ())

    print('SECOND:', second)
    assert second == ('term', 'b', (('text', r'les amis, \"coucou\".'),))

    print('THIRD:', third)
    expected = ('rule', ('term', 'obj', ('X',)), (
        ('term', 'rel', ('X', '_')),
        ('forall', 'rel', ('X', 'Y'), (('term', 'att', ('Y',)),)),
    ))
    print('EXPEC:', expected)
    assert third == expected

    print('FOURTH:', fourth)
    expected = ('rule', ('term', 'att', ('Y',)), (
        ('term', 'rel', ('_', 'Y')),
        ('forall', 'rel', ('X', 'Y'), (('term', 'obj', ('X',)),)),
    ))
    print('EXPECT:', expected)
    assert fourth == expected




import random
import pytest
from clyngor import asp_parsing


@pytest.fixture
def asp_source_and_graph():
    return """
    rel(a,c). rel(a,d). rel(b,d). rel(b,e).
    obj(X):- rel(X,_) ; rel(X,Y): att(Y).
    att(Y):- rel(_,Y) ; rel(X,Y): obj(X).
    :- not obj(X):obj(X).
    :- not att(Y):att(Y).
    """, {
        0: {4, 5}, 1: {4, 5}, 2: {4, 5}, 3: {4, 5},   # data nodes
        4: {5, 6},  # obj/1
        5: {4, 7},  # att/1
    }


def test_interdependency_graph(asp_source_and_graph):
    source, expected = asp_source_and_graph
    graph = asp_parsing.program_to_dependancy_graph(source)
    assert graph == expected



from clyngor import solve


# def test_time_limit_with_solutions():
    # """Sudoku yield thousands of solution in a second"""
    # answers = solve([], inline=SUDOKU, options='--time-limit=1')
    # nb_answer = sum(1 for answer in answers)
    # assert nb_answer > 1

# def test_time_limit_no_solutions():
    # """Queens do not yield any solution in a second
    # (at least on not so powerful machines)

    # This is tested because clingo do not raise the same error code
    # depending of whether a solution was found or not when stopped
    # because of the --time-limit option.

    # """
    # answers = solve([], inline=QUEENS, options='--time-limit=1')
    # nb_answer = sum(1 for answer in answers)
    # assert nb_answer == 0

# def test_no_time_limit_queens():
    # """Queens yield at least one solution when enough time is provided.

    # This is tested in order to prove that the time limit option works
    # properly in the `test_time_limit_no_solutions` function.

    # """
    # answers = solve([], inline=QUEENS, nb_model=1)
    # assert sum(1 for answer in answers) == 1


# QUEENS = """
# #const n = 200.
# n(1..n).

# q(X,Y) :- n(X), n(Y), not not q(X,Y).

        # c(r,X; c,Y) :- q(X,Y).
# not not c(r,N; c,N) :- n(N).

# n(r,X,Y-1,X,Y; c,X-1,Y,X,Y; d1,X-1,Y-1,X,Y;     d2,X-1,Y+1,X,Y      ) :- n(X), n(Y).
# c(r,N,0;       c,0,N;       d1,N-1,0; d1,0,N-1; d2,N-1,n+1; d2,0,N+1) :- n(N).

# c(C,XX,YY) :-     c(C,X,Y), n(C,X,Y,XX,YY), not q(XX,YY).
           # :- not c(C,X,Y), n(C,X,Y,XX,YY),     q(XX,YY).

# #show q/2.
# """


# SUDOKU = """
# val(1..9).
# border(1;4;7).

# % Only one number per square.
# 1 { x(X,Y,N) : val(N) } 1 :- val(X) ; val(Y).

# % Same numbers does not share row or column…
# 1 { x(X,Y,N) : val(X) } 1 :- val(N) ; val(Y).
# 1 { x(X,Y,N) : val(Y) } 1 :- val(N) ; val(X).

# % … nor boxes.
# 1 { x(X,Y,N) : val(X), val(Y), X1<=X, X<=X1+2, Y1<=Y, Y<=Y1+2 } 1 :- val(N) ; border(X1) ; border(Y1).
# """




import pytest
from clyngor.answers import Answers


@pytest.fixture
def simple_answers():
    return Answers((
        'a(0) b(1)',
        'c(2) d(3)',
        'e(4) f(5)',
        'g(6) h(7)',
    ))

@pytest.fixture
def noarg_answers():
    return Answers((
        '',
        'a',
        'b c',
        'd e f',
    ))


def test_multiple_tunning_no_arg(noarg_answers):
    answers = noarg_answers.no_arg
    assert next(answers) == frozenset()
    assert next(answers) == {'a'}
    assert next(answers) == {'b', 'c'}
    answers.by_predicate
    assert next(answers) == {'d': frozenset(), 'e': frozenset(), 'f': frozenset()}


def test_multiple_tunning(simple_answers):
    answers = simple_answers
    assert next(answers) == {('a', (0,)), ('b', (1,))}
    answers.by_predicate
    assert next(answers) == {'c': {(2,)}, 'd': {(3,)}}
    answers.first_arg_only
    assert next(answers) == {'e': {4}, 'f': {5}}
    answers.int_not_parsed
    assert next(answers) == {'g': {'6'}, 'h': {'7'}}


def test_tunning_first_arg_only(simple_answers):
    answers = simple_answers
    answers.first_arg_only
    assert next(answers) == {('a', 0), ('b', 1)}
    assert next(answers) == {('c', 2), ('d', 3)}
    assert next(answers) == {('e', 4), ('f', 5)}




import tempfile
from .test_api import asp_code  # fixture
from clyngor import ASP, utils


def test_save_in_file(asp_code):
    answers = frozenset(ASP(asp_code))
    with tempfile.NamedTemporaryFile('w', delete=False) as ofd:
        ofd.write('\n'.join(utils.answer_set_to_str(answer) for answer in answers))
        fname = ofd.name

    with open(fname) as ifd:
        read_answers = frozenset(
            frozenset(utils.answer_set_from_str(line))
            for line in ifd
        )

    assert read_answers == answers


def test_file_saving_api(asp_code):
    answers = frozenset(ASP(asp_code))
    fname = utils.save_answers_in_file(answers)
    read_answers = frozenset(utils.load_answers_from_file(fname))
    assert answers == read_answers


def test_file_saving_api_with_dict(asp_code):
    # create the read_answers from the dictionary repr of answer sets
    as_dict_answers = tuple(ASP(asp_code).by_predicate)
    fname = utils.save_answers_in_file(as_dict_answers)
    read_answers = frozenset(utils.load_answers_from_file(fname))

    # must be the same as regular repr of answer sets
    answers = frozenset(ASP(asp_code))
    assert answers == read_answers


def test_file_saving_api_with_as_pyasp(asp_code):
    # create the read_answers from the dictionary repr of answer sets
    as_dict_answers = tuple(ASP(asp_code).as_pyasp)
    fname = utils.save_answers_in_file(as_dict_answers)
    read_answers = frozenset(utils.load_answers_from_file(fname))

    # must be the same as regular repr of answer sets
    answers = frozenset(ASP(asp_code))
    assert answers == read_answers





import re
import itertools
import pypeg2 as pg


def parse_asp(source_code:str) -> tuple:
    return tuple(parsed_to_tuple(pg.parse(text=source_code, thing=Program, comment=comment_asp)))

def parsed_to_tuple(program) -> tuple:
    return program.as_tuple()



comment_asp = [re.compile(r'%\*.*\*%', flags=re.DOTALL), re.compile('%.*')]

Ident = re.compile(r'[a-z][a-zA-Z0-9_]*')
Variable = re.compile(r'(([A-Z][a-zA-Z0-9_]*)|(_))')


def to_tuple(x):
    return x.as_tuple() if hasattr(x, 'as_tuple') else x


class List(pg.List):
    """A pypeg2 list exposing a basic transformation routine"""
    flat = False

    def as_tuple(self):
        if type(self).flat:
            return tuple(itertools.chain.from_iterable(map(to_tuple, self)))
        else:
            return tuple(map(to_tuple, self))

class Number(List):
    grammar = re.compile(r'-?[0-9]+')
    def as_tuple(self):
        assert len(self) == 1
        return int(tuple(self)[0])

class Text(List):
    grammar = '"', re.compile(r'((\\")|([^"]))*'), '"'
    def as_tuple(self):
        assert len(self) == 1
        return ('text', tuple(self)[0])

class Litteral(List):
    grammar = [Number, Ident, Text, Variable]
    def as_tuple(self):
        assert len(self) == 1
        return to_tuple(tuple(self)[0])

class Arg(List):
    grammar = Litteral  # updated later
    def as_tuple(self):
        assert len(self) == 1
        return tuple(self)[0].as_tuple()
class Args(List):
    grammar = Arg, pg.maybe_some(',', Arg)
class MultArgs(List):
    grammar = Args, pg.maybe_some(';', Args)
    def as_tuple(self):
        if len(self) > 1:
            return ('disjunction', *tuple(map(to_tuple, self)))
        return tuple(map(to_tuple, tuple(self)[0]))

class UnamedTerm(List):
    grammar = '(', MultArgs, ')'
    def as_tuple(self):
        assert len(self) == 1
        children = tuple(self)[0].as_tuple()
        if children[0] == 'disjunction':
            return children
        return ('term', *children)
class NamedTerm(List):
    grammar =  Ident, pg.optional('(', MultArgs, ')')
    def as_tuple(self):
        children = tuple(self)
        assert len(children) in {1, 2}
        if len(children) == 1:
            return ('term', to_tuple(children[0]), ())
        return ('term', *map(to_tuple, children))
class Term(List):
    grammar = [UnamedTerm, NamedTerm]
    flat = True
Arg.grammar = [Term, Litteral]


# body constructions
class NotTerm(List):
    grammar = 'not', NamedTerm
    def as_tuple(self):
        type, pred, args = tuple(self)[0].as_tuple()
        assert type == 'term'
        return '¬' + type, pred, args
class ForAll(List):
    grammar = NamedTerm, ':', Term, pg.maybe_some(',', Term)
    def as_tuple(self):
        head, *conditions = self
        head = head.as_tuple()[1:]  # only the predicate and args
        conditions = tuple(sub.as_tuple() for sub in conditions)
        return ('forall', *head, conditions)
class NotForAll(List):
    grammar = 'not', ForAll
    def as_tuple(self):
        type, pred, args, conditions = tuple(self)[0].as_tuple()
        assert type == 'forall'
        return '¬' + type, pred, args, conditions
class Expression(List):
    grammar = [NotForAll, ForAll, NotTerm, Term]
    flat = True

class Selection(List):
    grammar = pg.optional(Number), '{', pg.some(Expression), '}', pg.optional(Number)
    def as_tuple(self):
        children = tuple(map(to_tuple, self))
        down, up, exprs = 0, None, ()
        assert len(children) in {1, 2, 3}
        print('CHILDS:', children)
        if isinstance(children[0], int):
            down = children[0]
        if isinstance(children[-1], int):
            up = children[-1]
        if up and down:
            assert len(children) == 3
            exprs = children[1]
        elif up:
            assert len(children) == 2
            exprs = children[0]
        elif down:
            assert len(children) == 2
            exprs = children[1]
        else:
            assert len(children) == 1
            exprs = children[0]

        return 'selection', down, up, (exprs,)


class Body(List):
    grammar = Expression, pg.maybe_some(';', Expression)
    def as_tuple(self):
        return tuple(sub.as_tuple() for sub in self)

class Head(List):
    grammar = [Selection, NamedTerm, (NamedTerm, pg.some(';', NamedTerm))]
    def as_tuple(self):
        assert len(self) == 1
        body = self[0]
        if ';' in body:
            return 'disjunction', body.as_tuple()
        return body.as_tuple()

class Constraint(List):
    grammar = ':-', Body
    def as_tuple(self):
        assert len(self) == 1
        body = self[0]
        return 'constraint', body.as_tuple()

class Rule(List):
    grammar = Head, ':-', Body
    def as_tuple(self):
        head, body = self
        return 'rule', head.as_tuple(), body.as_tuple()

class Program(List):
    grammar = pg.some([Constraint, Rule, Head], '.')



"""Routines and classes for Answer Set Programming source code parsing.

This parser is precise: it will carefully parse all asp source code
(except the comments).

For a more high level parser, see byline_parser.

"""

import arpeggio as ap

from clyngor.asp_parsing import asp_grammar, asp_grammar_comments
from clyngor.asp_parsing import alt_parse


def parse_asp_program_by_arpeggio(asp_source_code:str, do=None, have_comments:bool=True) -> tuple:
    parser = ap.ParserPython(asp_grammar(), asp_grammar_comments if have_comments else None)
    parse_tree = parser.parse(asp_source_code)
    return ap.visit_parse_tree(parse_tree, visitor=do or CodeAsTuple())

def parse_asp_program_by_pypeg(asp_source_code:str, do=None, have_comments:bool=True) -> tuple:
    return alt_parse.parse_asp(asp_source_code)

# handles multiline comments, but is slower
parse_asp_program = parse_asp_program_by_pypeg
parse_asp_program = parse_asp_program_by_arpeggio


class CodeAsTuple(ap.PTNodeVisitor):
    def __init__(self, int_builder:type=int):
        super().__init__()
        self._int_builder = int_builder

    def visit_number(self, node, children):
        return self._int_builder(node.value)

    def visit_args(self, node, children):
        return tuple(children)

    def visit_multargs(self, node, children):
        if len(children) > 1:
            return ('disjunction', *children)
        return tuple(children)


    def visit_text(self, node, children):
        if len(children) == 0:  # empty text
            text = ''
        else:
            assert len(children) == 1
            text = tuple(children)[0]
        return 'text', str(text)

    def visit_ident(self, node, children):
        return str(node.value)

    def visit_unnamedterm(self, node, children):
        # children is a list of element in disjunction
        assert len(children) == 1
        args = children[0]
        if args[0] == 'disjunction':
            return args
        return 'term', None, tuple(children)

    def visit_namedterm(self, node, children):
        assert len(children) in {1, 2}
        if len(children) == 1:  # just an ident
            predicate, args = children[0], ()
        else:  # ident + args
            predicate, args = children
            if len(args) > 1:  # disjunction
                args = ('disjunction', *args)
            else:
                assert len(args) == 1
                args = args[0]
        return 'term', predicate, tuple(args)

    def visit_not_term(self, node, children):
        assert len(children) == 1
        type, predicate, args = children[0]
        return '¬' + type, predicate, args

    def visit_term(self, node, children):
        assert len(children) == 1
        type, predicate, args = children[0]
        return type, predicate, args

    def visit_not_forall(self, node, children):
        atom, *conditions = children
        assert len(atom) == 3
        assert atom[0] == 'term'
        return '¬forall', atom[1], atom[2], tuple(conditions)

    def visit_forall(self, node, children):
        atom, *conditions = children
        assert len(atom) == 3
        assert atom[0] == 'term'
        return 'forall', atom[1], atom[2], tuple(conditions)

    def visit_expression(self, node, children):
        return tuple(children)

    def visit_body(self, node, children):
        predicate, *args = children
        for child in children:
            assert len(child) == 1
        return tuple(child[0] for child in children)

    def visit_constraint(self, node, children):
        if len(children) == 1:
            children = children[0]
        else:
            raise ValueError("Constraint body has not 1 value, but {}: {}".format(len(children), children))
        return 'constraint', tuple(children)


    def visit_selection(self, node, children):
        downum, upnum, body = 0, None, ()
        if len(children) == 3:
            downum, expr, upnum = children
        elif len(children) == 1:
            expr = children[0]
        else:
            assert len(children) == 2
            if isinstance(children[0], tuple):  # second number given
                expr, upnum = children
            else:  # first number given
                upnum, expr = children
        return 'selection', downum, upnum, expr


    def visit_head(self, node, children):
        predicate, *args = children
        assert len(children) == 1
        return children[0]

    def visit_rule(self, node, children):
        assert len(children) == 2
        head, body = children
        return ('rule', tuple(head), tuple(body))

    def visit_multirule(self, node, children):
        assert len(children) > 2  # at least two heads and a body
        head, body = children[:-1], children[-1]
        return ('multirule', tuple(head), tuple(body))


    def visit_arityterm(self, node, children):
        return 'arity', children[0], self._int_builder(children[1])

    def visit_meta_show(self, node, children):
        assert len(children) == 1
        type, *child = children[0]
        return ('show/n' if type == 'arity' else 'show', *child)

    def visit_meta_const(self, node, children):
        assert len(children) == 2
        return ('const', *tuple(children))


    def visit_program(self, node, children):
        for rule in children:
            yield rule



"""Routines to manipulate ASP source code as a graph.

The following data structure are used here.

Dependancy graph -- mapping {rule: {dependancies}} where keys are rules
                    and values are iterable of rules yielding atoms used by key.
atoms endpoints -- tuple (source, target) where source and target are mappings
                   of atoms to the rules that yield them, or the rules
                   that use them, respectively.
program -- a tuple representation of an ASP source code, as parsed by
           asp_parsing.parser module using CodeAsTuple visitor.
           In this representation, rules should be ordered, or the indexes
           of other data structures may be wrong.

Note that atoms are identified by their predicate.
This behavior may change in the future to include arity (number of parameter),
which is much more accurate.

"""
import itertools
from collections import defaultdict
from clyngor.asp_parsing.precise_parser import parse_asp_program, CodeAsTuple


def program_to_endpoints(program:str or tuple, node_as_index:bool=True) -> dict:
    """Return the atoms endpoints computed from given parsed ASP program.

    program -- either a source code as string or a parsed one
    node_as_index -- return nodes as index of rules in the source code
                     instead of rules themselves

    """
    if isinstance(program, str):
        program = tuple(parse_asp_program(program, do=CodeAsTuple()))
    atom_source = defaultdict(set)  # predicate -> rules yielding it
    atom_target = defaultdict(set)  # predicate -> rules using it

    # explore the full structure to populate sources and targets
    for idx, (type, *data) in enumerate(program):
        if type == 'rule':
            head, body = data
            atom_source[head[1]].add(idx)
            for (subtype, *subdata) in body:
                if subtype in {'term', '¬term'}:
                    atom_target[subdata[0]].add(idx)
                elif subtype in {'forall', '¬forall'}:
                    atom_target[subdata[0]].add(idx)
                    for (subsubtype, *subsubdata) in subdata[2]:
                        if subsubtype == 'term':
                            atom_target[subsubdata[0]].add(idx)
                        else:
                            assert False, "Type {} is unexpected in forall body".format(subtype)
                else:
                    assert False, "Type {} is unexpected in rule body".format(subtype)

        elif type == 'term':
            atom_source[data[0]].add(idx)
        elif type == 'selection':
            for (subtype, *subdata) in data[2]:
                if subtype in {'term', '¬term', 'forall', '¬forall'}:
                    atom_source[subdata[0]].add(idx)
                else:
                    assert False, "Type {} is unexpected in selection".format(subtype)
        elif type == 'constraint':
            for (subtype, *subdata) in data[0]:
                if subtype in {'term', '¬term', 'forall', '¬forall'}:
                    atom_target[subdata[0]].add(idx)
                else:
                    assert False, "Type {} is unexpected in selection".format(subtype)
        else:
            assert False, "Type {} is unexpected at first level".format(type)

    if node_as_index:
        return dict(atom_source), dict(atom_target)
    else:  # use the parsed rules themselves
        return tuple(
            {program[pred]: frozenset(program[succ] for succ in succs)
             for pred, succs in graph.items()}
            for graph in (atom_source, atom_target)
        )


def atoms_endpoints_to_dependancy_graph(atom_source:dict, atom_target:dict) -> dict:
    """Return the dependancy graph"""
    graph = defaultdict(set)
    for predicate in set(atom_source.keys()) | set(atom_target.keys()):
        for source, target in itertools.product(atom_source[predicate], atom_target[predicate]):
            graph[source].add(target)
    return dict(graph)


def program_to_dependancy_graph(program:str or tuple,
                                node_as_index:bool=True,
                                have_comments:bool=True) -> dict:
    """Most high level function: compute dependancy graph from the source code
    directly"""
    if isinstance(program, str):
        program = tuple(parse_asp_program(program, do=CodeAsTuple(), have_comments=have_comments))
    endpoints = program_to_endpoints(program, node_as_index)
    return atoms_endpoints_to_dependancy_graph(*endpoints)



"""Routines and classes for Answer Set Programming source code parsing.

"""

import re
import itertools
from collections import defaultdict
import arpeggio as ap


def parse_asp_program(asp_source_code:str, do=None) -> tuple:
    parse_tree = ap.ParserPython(asp_grammar()).parse(asp_source_code)
    return ap.visit_parse_tree(parse_tree, visitor=do or CodeAsTuple())


def rules_interdependency_graph(program:str or tuple,
                                node_as_index:bool=True) -> dict:
    """Return the directed graph as {predecessor: {successors}} where nodes
    are rules in the input ASP source code, and source of a link
    is the rule yielding atom used by the target.

    program -- either a source code as string or a parsed one
    node_as_index -- return nodes as index of rules in the source code
                     instead of rules themselves

    """
    if isinstance(program, str):
        program = tuple(parse_asp_program(program, do=CodeAsTuple()))
    else:
        program
    atom_source = defaultdict(set)  # predicate -> rules yielding it
    atom_target = defaultdict(set)  # predicate -> rules using it

    # explore the full structure to populate sources and targets
    for idx, (type, *data) in enumerate(program):
        if type == 'rule':
            atom_source[data[0]].add(idx)
            for (subtype, *subdata) in data[2]:
                if subtype in {'term', '¬term'}:
                    atom_target[subdata[0]].add(idx)
                elif subtype in {'forall', '¬forall'}:
                    atom_target[subdata[0]].add(idx)
                    for (subsubtype, *subsubdata) in subdata[2]:
                        if subsubtype == 'term':
                            atom_target[subsubdata[0]].add(idx)
                        else:
                            assert False, "Type {} is unexpected in forall body".format(subtype)
                else:
                    assert False, "Type {} is unexpected in rule body".format(subtype)

        elif type == 'term':
            atom_source[data[0]].add(idx)
        elif type == 'selection':
            for (subtype, *subdata) in data[2]:
                if subtype in {'term', '¬term', 'forall', '¬forall'}:
                    atom_source[subdata[0]].add(idx)
                else:
                    assert False, "Type {} is unexpected in selection".format(subtype)
        elif type == 'constraint':
            for (subtype, *subdata) in data[0]:
                if subtype in {'term', '¬term', 'forall', '¬forall'}:
                    atom_source[subdata[0]].add(idx)
                else:
                    assert False, "Type {} is unexpected in selection".format(subtype)
        else:
            assert False, "Type {} is unexpected at first level".format(type)

    # build the dependancy graph
    graph = defaultdict(set)
    for predicate in set(atom_source.keys()) | set(atom_target.keys()):
        for source, target in itertools.product(atom_source[predicate], atom_target[predicate]):
            graph[source].add(target)
    if node_as_index:
        return dict(graph)
    else:  # use the parsed rules themselves
        return {program[pred]: frozenset(program[succ] for succ in succs)
                for pred, succs in graph.items()}



def asp_grammar():
    """Implement the Answer Set Programming grammar.
    """
    def ident():      return ap.RegExMatch(r'[a-z][a-zA-Z0-9_]*')
    def number():     return ap.RegExMatch(r'-?[0-9]+')
    def text():       return '"', ap.RegExMatch(r'((\\")|([^"]))*'), '"'
    def litteral():   return [ident, text, number]
    def variable():   return ap.RegExMatch(r'(([A-Z][a-zA-Z0-9_]*)|(_))')
    def subterm():    return [(ident, ap.Optional("(", args, ")")), litteral, variable]
    def args():       return subterm, ap.ZeroOrMore(',', subterm)
    def term():       return ap.Optional('not'), ident, ap.Optional("(", args, ")")
    def forall():     return term, ':', term, ap.ZeroOrMore(',', term)
    def expression(): return [forall, term]
    def body():       return expression, ap.ZeroOrMore(';', expression)
    def constraint(): return ':-', body
    def head():       return term, ap.ZeroOrMore(';', term)
    def rule():       return head, ':-', body
    def selection_nobody():  return ap.Optional(number), '{', ap.OneOrMore(expression), '}', ap.Optional(number)
    def selection_body():  return ap.Optional(number), '{', ap.OneOrMore(expression), '}', ap.Optional(number), ':-', body
    def selection():  return [selection_body, selection_nobody]
    def instruction():return [selection, constraint, rule, term]
    def program():    return ap.OneOrMore(instruction, '.')
    return program


class CodeAsTuple(ap.PTNodeVisitor):
    def __init__(self, int_builder:type=int):
        super().__init__()
        self._int_builder = int_builder

    def visit_number(self, node, children):
        return self._int_builder(node.value)

    def visit_args(self, node, children):
        return tuple(children)

    def visit_text(self, node, children):
        text = tuple(children)[0]
        return 'text', str(text)

    def visit_ident(self, node, children):
        return str(node.value)

    def visit_term(self, node, children):
        if children[0] == 'not':
            type = '¬term'
            predicate, *args = children[1:]
        else:
            type = 'term'
            predicate, *args = children
        assert len(args) in {0, 1}
        args = args[0] if len(args) == 1 else args
        return type, predicate, tuple(args)

    def visit_forall(self, node, children):
        atom, *conditions = children
        assert len(atom) == 3
        assert atom[0] in {'term', '¬term'}
        type = '¬forall' if atom[0] == '¬term' else 'forall'
        return type, atom[1], atom[2], tuple(conditions)

    def visit_expression(self, node, children):
        return tuple(children)

    def visit_body(self, node, children):
        predicate, *args = children
        for child in children:
            assert len(child) == 1
        return tuple(child[0] for child in children)

    def visit_constraint(self, node, children):
        if len(children) == 1:
            children = children[0]
        else:
            raise ValueError("Constraint body has not 1 value, but {}: {}".format(len(children), children))
        return 'constraint', tuple(children)


    def visit_selection_nobody(self, node, children):
        downum, upnum, body = 0, None, ()
        if len(children) == 3:
            downum, expr, upnum = children
        elif len(children) == 1:
            expr = children[0]
        else:
            assert len(children) == 2
            if isinstance(children[0], tuple):  # second number given
                expr, upnum = children
            else:  # first number given
                upnum, expr = children
        return 'selection', downum, upnum, expr, body


    def visit_selection_body(self, node, children):
        downum, upnum, body = 0, None, ()
        if len(children) == 4:
            downum, expr, upnum, body = children
        elif len(children) == 2:  # no numbers
            expr, body = children
        else:  # one number is given ; which one ?
            assert len(children) == 3
            tuples = tuple(isinstance(child, tuple) for child in children)
            assert any(tuples)
            if tuples[0]:  # the second number is given
                assert not tuples[1]
                expr, upnum, body = children
            elif tuples[1]:  # the first number is given
                downum, expr, body = children
            else:  # no number given WTF
                assert False, "algorithm is false."
            # determine if body is given
        return 'selection', downum, upnum, expr, body

    def visit_head(self, node, children):
        predicate, *args = children
        return children

    def visit_rule(self, node, children):
        head, *body = children
        if len(head) == 1:  # general case: only one production
            head = head[0][1:]  # remove the 'term'
            type = 'rule'
        else:
            type = 'multirule'
        if len(body) == 1:
            body = body[0]
        else:
            raise ValueError("Body has not 1 value, but {}: {}".format(len(body), body))
        return (type, *tuple(head), tuple(body))

    def visit_program(self, node, children):
        for rule in children:
            yield rule



import os
from setuptools   import setup, find_packages
from pip.req      import parse_requirements
from pip.download import PipSession


# access to the file at the package top level (like README)
def path_to(filename):
    return os.path.join(os.path.dirname(__file__), filename)
# parse_requirements() returns generator of pip.req.InstallRequirement objects
install_reqs = parse_requirements(path_to('requirements.txt'), session=PipSession())
reqs = [str(ir.req) for ir in install_reqs]


setup(
    name = 'clyngor',
    version = '0.0.13',
    packages = find_packages(),
    include_package_data = True,  # read the MANIFEST.in file
    install_requires = reqs,

    author = "lucas bourneuf",
    author_email = "lucas.bourneuf@laposte.net",
    description = "Python wrapper around Clingo/Answer Set Programming",
    long_description = open(path_to('README.mkd')).read(),
    keywords = "ASP clingo wrapper",
    url = "https://github.com/aluriak/clyngor",

    classifiers = [
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: GNU General Public License (GPL)",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: ASP",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
)



"""Routines for printings of parsed ASP programs.

"""


class BaseVisitor:
    """Base class for user-defined visitor over ASP source code"""

    def term(self, predicate:str, args:iter, not_:bool):
        raise NotImplementedError()

    def forall(self, predicate:str, args:iter, conditions:iter, not_:bool):
        raise NotImplementedError()

    def selection(self, down:int, up:int, generateds:iter):
        raise NotImplementedError()

    def rule(self, head:'term', body:iter):
        raise NotImplementedError()

    def multirule(self, head:iter, body:iter):
        raise NotImplementedError()

    def constraint(self, body:iter):
        raise NotImplementedError()

    def disjunction(self, items:iter):
        raise NotImplementedError()

    def litteral(self, item:str or int):
        raise NotImplementedError()

    def program(self, rules:iter):
        raise NotImplementedError()

    @classmethod
    def walk_on(cls, program:tuple, **kwargs):
        """Return the visitor results of its walk on given program"""
        visitor = cls(**kwargs)
        return self.program(map(visitor.visite, program))

    def visite(self, obj):
        type = obj[0]
        if type == 'rule':
            assert len(obj) == 3
            head, body = obj[1:]
            return self.visite(head) + rule_marker + body_sep.join(map(self.visite, body))
        if type == 'multirule':
            head, body = obj[1:]
            assert len(obj) == 3
            head = map(self.visite, head)
            body = map(self.visite, body)
            return disjunction_sep.join(head) + rule_marker + body_sep.join(body)
        if type == 'constraint':
            assert len(obj) == 2
            body = map(self.visite, obj[1])
            return rule_marker + body_sep.join(body)
        if type in {'term', '¬term'}:
            name, args = obj[1:]
            args = map(self.visite, args)
            not_ = ('not ' if type.startswith('¬') else '')
            return not_ + str_term(name, args)
        if type in {'forall', '¬forall'}:
            assert len(obj) == 4
            name, args, conditions = obj[1:]
            not_ = ('not ' if type.startswith('¬') else '')
            return not_ + str_term(name, map(self.visite, args)) + ':' + conditions_sep.join(map(self.visite, conditions))
        if type == 'disjunction':
            raise NotImplementedError()
        if type in 'selection':
            assert len(obj) == 4
            down, up, generateds = obj[1:]
            return str_selection(down, up, map(self.visite, generateds))
        # now here, it is a litteral
        return args_sep.join(map(str, obj))




def to_asp_source_code(program:tuple, rule_marker:str=':- ', body_sep:str=' ; ',
                       conditions_sep:str=', ', args_sep:str=',',
                       disjunction_sep:str=';', rule_end:str='.',
                       selection_always_show_numbers:bool=False) -> iter:
    """Yield a ASP compliant string implementing each rule in program.

    """
    def str_term(pred, args:iter) -> str:
        args = tuple(args)
        return '{}({})'.format(pred, args_sep.join(args)) if args else pred
    def str_selection(down:int, up:int, conditions:iter) -> str:
        out = ''
        if selection_always_show_numbers or down != 0:
            out += str(down) + ' '
        out += '{ ' + ';'.join(conditions) + ' }'
        if selection_always_show_numbers or up != None:
            out += ' ' + str(up)
        return out
    def str_disjunction(args:iter) -> str:
        return disjunction_sep.join(args)


    def stringifier(obj) -> callable:
        type = obj[0]
        if type == 'rule':
            assert len(obj) == 3
            head, body = obj[1:]
            return stringifier(head) + rule_marker + body_sep.join(map(stringifier, body))
        if type == 'multirule':
            head, body = obj[1:]
            assert len(obj) == 3
            head = map(stringifier, head)
            body = map(stringifier, body)
            return disjunction_sep.join(head) + rule_marker + body_sep.join(body)
        if type == 'constraint':
            assert len(obj) == 2
            body = map(stringifier, obj[1])
            return rule_marker + body_sep.join(body)
        if type in {'term', '¬term'}:
            name, args = obj[1:]
            args = map(stringifier, args)
            not_ = ('not ' if type.startswith('¬') else '')
            return not_ + str_term(name, args)
        if type in {'forall', '¬forall'}:
            assert len(obj) == 4
            name, args, conditions = obj[1:]
            not_ = ('not ' if type.startswith('¬') else '')
            return not_ + str_term(name, map(stringifier, args)) + ':' + conditions_sep.join(map(stringifier, conditions))
        if type == 'disjunction':
            raise NotImplementedError()
        if type in 'selection':
            assert len(obj) == 4
            down, up, generateds = obj[1:]
            return str_selection(down, up, map(stringifier, generateds))
        # now here, it is a litteral
        return args_sep.join(map(str, obj))

    for rule in program:
        yield stringifier(rule) + rule_end


def to_human_text(program:tuple) -> str:
    """Return describing the program for humans.

    """




from pprint import pprint
import pytest
from clyngor import ASP
from clyngor.asp_parsing import byline_parser
from clyngor.asp_parsing.byline_parser import SourceBlock


def test_solving():
    """Prove that the program is valid"""
    answers = tuple(ASP(ASP_SOURCE).by_predicate)
    assert len(answers) == 1
    for idx, answer in enumerate(answers):
        assert len(answer['p']) == 4
        assert len(answer['rel']) == 4
        assert answer['rel'] == {('a', 'b'), ('a', 'c'), ('a', 'd'), ('a', 'e')}


def test_byline_parsing():
    parsed = tuple(byline_parser.parse(ASP_SOURCE))
    pprint(parsed)
    assert parsed == (
        (0, 0, 'empty'),
        (1, 1, 'comment', (1, 19), ' this is a header'),
        (2, 4, 'multicomment', (20, 45), '\na multiline one %%*\n'),
        (5, 5, 'empty'),
        (6, 6, 'comment', (47, 72), ' a comment before a rule'),
        (7, 7, 'code', (73, 89), 'rel(a,(b;c;d;e))'),
        (7, 7, 'end', (89, 90), '.'),
        (8, 8, 'empty'),
        (9, 11, 'multicomment', (92, 131), ' a multiline comment\nbefore a rule\n'),
        (12, 12, 'code', (132, 147), 'p(X):- rel(a,X)'),
        (12, 12, 'end', (147, 148), '.'),
        (13, 13, 'comment', (149, 167), ' and after a rule'),
        (14, 14, 'empty'),
        (15, 15, 'comment', (169, 177), ' footer'),
        (16, 16, 'comment', (178, 189), '#show p/1.'),
        (17, 17, 'comment', (190, 203), '#show rel/2.'),
    )


def test_byline_rebuild():
    clusters = byline_parser.clusterize_from_source(ASP_SOURCE)
    rebuilt = byline_parser.rebuild_clusters(clusters)
    assert rebuilt == ASP_SOURCE.strip()


def test_line_ends():
    indexes = tuple(byline_parser.lines_end('\na\nb\ncde\n\n'))
    assert indexes == (0, 2, 4, 8, 9)


def test_clusterize():
    clusters = tuple(byline_parser.clusterize(byline_parser.parse(ASP_SOURCE)))
    print(clusters)
    clusters_from_source = tuple(byline_parser.clusterize_from_source(ASP_SOURCE))
    assert clusters == clusters_from_source
    expected = (
        SourceBlock(False, (
            (1, 1, 'comment', (1, 19), ' this is a header'),
            (2, 4, 'multicomment', (20, 45), '\na multiline one %%*\n')
        ), 1, 4),
        SourceBlock(True, (
            (6, 6, 'comment', (47, 72), ' a comment before a rule'),
            (7, 7, 'code', (73, 89), 'rel(a,(b;c;d;e))'),
            (7, 7, 'end', (89, 90), '.')
        ), 6, 7),
        SourceBlock(True, (
            (9, 11, 'multicomment', (92, 131), ' a multiline comment\nbefore a rule\n'),
            (12, 12, 'code', (132, 147), 'p(X):- rel(a,X)'),
            (12, 12, 'end', (147, 148), '.'),
            (13, 13, 'comment', (149, 167), ' and after a rule')
        ), 9, 13),
        SourceBlock(False, (
            (15, 15, 'comment', (169, 177), ' footer'),
            (16, 16, 'comment', (178, 189), '#show p/1.'),
            (17, 17, 'comment', (190, 203), '#show rel/2.')
        ), 15, 17)
    )
    assert clusters == expected


def test_clusterize_over_blocks():
    parsed = tuple(byline_parser.parse(ASP_SOURCE_RULE_OVER_BLOCKS))
    pprint(parsed)
    clusters = tuple(byline_parser.clusterize(parsed))
    clusters_from_source = tuple(byline_parser.clusterize_from_source(ASP_SOURCE_RULE_OVER_BLOCKS))
    assert clusters == clusters_from_source
    pprint(clusters)
    expected = (
        SourceBlock(True, (
            (1, 1, 'comment', (1, 12), ' comment 1'),
            (2, 2, 'code', (13, 19), 'rel(a)'),
            (3, 3, 'comment', (20, 31), ' comment 2'),
            (5, 5, 'comment', (33, 44), ' comment 3'),
            (6, 6, 'end', (45, 46), '.'),
            (7, 7, 'comment', (47, 58), ' comment 4'),
        ), 1, 7),
    )
    assert clusters == expected


ASP_SOURCE = """
% this is a header
%*
a multiline one %%*
*%

% a comment before a rule
rel(a,(b;c;d;e)).

%* a multiline comment
before a rule
*%
p(X):- rel(a,X).
% and after a rule

% footer
%#show p/1.
%#show rel/2.
"""


ASP_SOURCE_RULE_OVER_BLOCKS = """
% comment 1
rel(a)
% comment 2

% comment 3
.
% comment 4
"""
